from abc import ABC
import gym
from gym import spaces
from numpy import ndarray
from typing import Any, Generic, Tuple
from ..prelude import Action, State


class EnvSpec:
    def __init__(
            self,
            state_dim: Tuple[int, ...],
            action_space: gym.Space,
            use_reward_monitor: bool = False
    ) -> None:
        """Properties which are common both in EnvExt and ParallelEnv
        """
        self.state_dim = state_dim
        self.action_space = action_space
        self.use_reward_monitor = use_reward_monitor
        if isinstance(action_space, spaces.Discrete):
            self.action_dim = action_space.n
        elif isinstance(action_space, spaces.Box):
            if len(action_space.shape) != 1:
                raise RuntimeError('Box space with shape >= 2 is not supportd')
            self.action_dim = action_space.shape[0]
        else:
            raise RuntimeError('{} is not supported'.format(type(action_space)))

    def random_action(self) -> Action:
        return self.action_space.sample()

    def is_discrete(self) -> bool:
        return isinstance(self.action_space, spaces.Discrete)


class EnvExt(gym.Env, ABC, Generic[Action, State]):
    def __init__(self, env: gym.Env) -> None:
        self._env = env
        self.spec = EnvSpec(self._env.observation_space.shape, self._env.action_space)

    def close(self):
        """
        Inherited from gym.Env.
        """
        self._env.close

    def reset(self) -> State:
        """
        Inherited from gym.Env.
        """
        return self._env.reset()

    def render(self, mode: str = 'human') -> None:
        """
        Inherited from gym.Env.
        """
        self._env.render(mode=mode)

    def seed(self, seed: int) -> None:
        """
        Inherited from gym.Env.
        """
        self._env.seed(seed)

    def step(self, action: Action) -> Tuple[State, float, bool, Any]:
        """
        Inherited from gym.Env.
        """
        return self._env.step(action)

    def step_and_reset(self, action: Action) -> Tuple[State, float, bool, Any]:
        state, reward, done, info = self.step(action)
        if done:
            state = self.reset()
        return state, reward, done, info

    @property
    def unwrapped(self) -> gym.Env:
        """
        Inherited from gym.Env.
        """
        return self._env.unwrapped

    @property
    def action_dim(self) -> int:
        """
        Extended method.
        Returns a ndim of action space.
        """
        return self.spec.action_dim

    @property
    def state_dim(self) -> Tuple[int, ...]:
        """
        Extended method.
        Returns a shape of observation space.
        """
        return self.spec.state_dim

    @property
    def use_reward_monitor(self) -> bool:
        """Atari wrappers need RewardMonitor for evaluation.
        """
        return self.spec.use_reward_monitor

    @property
    def observation_space(self) -> gym.Space:
        return self._env.observation_space

    @property
    def action_space(self) -> gym.Space:
        return self._env.action_space

    @staticmethod
    def extract(state: State) -> ndarray:
        """
        Extended method.
        Convert state to ndarray.
        It's useful for the cases where numpy.ndarray representation is too large to
        throw it to replay buffer directly.
        """
        return state

    def save_history(self, file_name: str) -> None:
        """
        Extended method.
        Save agent's action history to file.
        """
        pass

    def __repr__(self) -> str:
        return 'EnvExt({})'.format(self._env)



import click
from typing import Optional
from .utils.log import ExperimentLog


@click.command()
@click.option('--log-dir', type=str)
@click.option('--vi-mode', is_flag=True)
def ipython(log_dir: Optional[str], vi_mode: bool) -> None:
    if log_dir is not None:
        log = ExperimentLog(log_dir)  # noqa
    else:
        open_log = ExperimentLog  # noqa
    try:
        from ptpython.ipython import embed
        from matplotlib import pyplot as plt  # noqa
        import rainy  # noqa
        del log_dir
        embed(vi_mode=vi_mode)
    except ImportError:
        print("To use ipython mode, install ipython and ptpython first.")


if __name__ == '__main__':
    ipython()



from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar

ReplayFeed = TypeVar('ReplayFeed')


class ReplayBuffer(ABC, Generic[ReplayFeed]):
    def __init__(self, feed: ReplayFeed) -> None:
        self.feed = feed

    @abstractmethod
    def append(self, *args) -> None:
        pass

    @abstractmethod
    def sample(self, batch_size: int) -> List[ReplayFeed]:
        pass

    @abstractmethod
    def __len__(self):
        pass




"""Dummy environment for testing
"""
from enum import Enum
from gym.spaces import Discrete
import numpy as np
from typing import Tuple
from .ext import EnvExt, EnvSpec

ACTION_DIM = 10


class State(Enum):
    START = 0
    ROAD1 = 1
    ROAD2 = 2
    DEATH = 3
    GOAL = 4

    def is_end(self) -> bool:
        return self == State.DEATH or self == State.GOAL

    def to_array(self, dim: tuple) -> np.array:
        return np.repeat(float(self.value) + 1.0, np.prod(dim)).reshape(dim)


class DummyEnv(EnvExt):
    def __init__(self, array_dim: Tuple[int, ...] = (16, 16), flatten: bool = False) -> None:
        self.state = State.START
        self.transition = [
            [0.0, 0.7, 0.3, 0.0, 0.0],
            [0.0, 0.0, 0.8, 0.2, 0.0],
            [0.0, 0.0, 0.0, 0.4, 0.6],
            [0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
        ]
        self.rewards = [0., 0., 0., -10., 20.]
        self.array_dim = array_dim
        self.flatten = flatten

    def reset(self) -> State:
        self.state = State.START
        return self.state

    def step(self, _action) -> Tuple[State, float, bool, dict]:
        prob = np.asarray(self.transition[self.state.value])
        self.state = State(np.random.choice(np.arange(5), 1, p=prob))
        return self.state, self.rewards[self.state.value], self.state.is_end(), {}

    @property
    def spec(self) -> EnvSpec:
        if self.flatten:
            dim = (np.prod(self.array_dim),)
        else:
            dim = self.array_dim
        return EnvSpec(dim, Discrete(ACTION_DIM))

    def seed(self, int) -> None:
        pass

    def close(self) -> None:
        pass

    def extract(self, state: State) -> np.ndarray:
        res = state.to_array(self.array_dim)
        if self.flatten:
            return res.flatten()
        else:
            return res


class DummyEnvDeterministic(DummyEnv):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.transition = [
            [0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
        ]



import torch
from .a2c import A2cAgent
from ..config import Config
from ..net import Policy


class AcktrAgent(A2cAgent):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        if self.net.is_recurrent:
            raise NotImplementedError('K-FAC for RNN is not implemented!')
        self.precond = config.preconditioner(self.net)

    def _pre_backward(self, policy: Policy, value: torch.Tensor) -> None:
        """Calculate emprical fisher loss
        """
        self.net.zero_grad()
        policy_fisher_loss = -policy.log_prob().mean()
        sample_value = torch.randn_like(value) + value.detach()
        value_fisher_loss = -(value - sample_value).pow(2).mean()
        fisher_loss = policy_fisher_loss + value_fisher_loss
        with self.precond.save_grad():
            fisher_loss.backward(retain_graph=True)

    def _step_optimizer(self) -> None:
        """Approximates F^-1∇h and apply it.
        """
        self.precond.step()
        self.optimizer.step()
        self.lr_cooler.lr_decay(self.optimizer)
